<?php

/**
 * Default implementation using database table.
 */
class Oox_Lock_Database implements Oox_Lock_Interface
{
  /**
   * @see Oox_Lock_Interface::acquire()
   */
  public function acquire($name, $lifetime = LIFETIME_DEFAULT, $group = Oox_Lock_Interface::DEFAULT_GROUP, $forUser = FALSE, $token = NULL) {
    $time = time();

    // Aptempt to find if this semaphore already been locked.
    $lock = db_select('semaphore_oox', 's')
      ->fields('s')
      ->condition('expire', $time, '>')
      ->condition('name', $name)
      ->execute()
      ->fetchObject();

    if ($lock) {
      // Success in case of identical tokens.
      if ($token && $lock->token == $token) {
        return TRUE;
      }
      else {
        throw new LockAlreadyAcquiredException($lock);
      }
    }
    else {
      // Avoid collisions when an expired semaphore still exists with the
      // same name.
      db_delete('semaphore_oox')
        ->condition('name', $name)
        ->execute();
    }

    // Get user information.
    if ($forUser) {
      global $user;
      $uid = $user->uid;
    }
    else {
      $uid = 0;
    }

    // If none found, then aptempt the acquisition transaction.
    $transaction = db_transaction();
    try {
      db_insert('semaphore_oox')->fields(array(
        'name' => $name,
        'value' => $group,
        'expire' => $time + $lifetime,
        'uid' => $uid,
        'token' => $token,
      ))->execute();
      // Force transaction to commit.
      unset($transaction);
    }
    // Check for transaction error.
    catch (PDOException $e) {
      $transaction->rollback();
      // Specific case, that should not happen that often, two transactions
      // have been overlapping each other. We should fetch back the current
      // lock and throw the exception we the right user.
      $lock = db_select('semaphore_oox', 's')
        ->fields('s')
        ->condition('expire', $time, '>')
        ->condition('name', $name)
        ->execute()
        ->fetchObject();
      throw new LockAlreadyAcquiredException($lock);
    }
  }

  /**
   * @see Oox_Lock_Interface::release()
   */
  public function release($name) {
    $transaction = db_transaction();
    try {
      db_delete('semaphore_oox')
        ->condition('name', $name)
        ->execute();
      // Force transaction to commit.
      unset($transaction);
    }
    // Check for transaction error.
    catch (PDOException $e) {
      $transaction->rollback();
      // This is a real error that should not happen, log why.
      watchdog_exception('oox_lock', $e);
    }
  }

  /**
   * @see Oox_Lock_Interface::getLock()
   */
  public function getLock($name, $group = NULL, $uid = NULL, $token = NULL) {
    // Prepare query.
    $select = db_select('semaphore_oox')
      ->fields('s')
      ->condition('expire', time(), '>')
      ->condition('name', $name);
    // Check for given group
    if ($group) {
      $select->condition('value', $group);
    }
    if ($uid) {
      $select->condition('uid', $uid);
    }
    if ($token) {
      $select->condition('token', $token);
    }
    return $select->execute()->fetchObject();
  }

  /**
   * @see Oox_Lock_Interface::wipeOutLocks()
   */
  public function wipeOutLocks($uid = 0, $group = NULL, $token = NULL) {
    // Prepare query.
    $delete = db_delete('semaphore_oox');
    $all = TRUE;
    // Check for given group
    if ($group) {
      $delete->condition('value', $group);
      $all = FALSE;
    }
    if ($uid) {
      $delete->condition('uid', $uid);
      $all = FALSE;
    }
    if ($token) {
      $delete->condition('token', $token);
      $all = FALSE;
    }
    // May be this is a maintenance call.
    if ($all) {
      $delete->condition('expire', time(), '<');
    }
    // Do the delete inside a transaction, it could overlap with a lock
    // acquire transaction.
    $transaction = db_transaction();
    try {
      $delete->execute();
      // Force transaction to commit.
      unset($transaction);
    }
    // Check for transaction error.
    catch (PDOException $e) {
      $transaction->rollback();
      // This is a real error that should not happen, log why.
      watchdog_exception('oox_lock', $e);
    }
  }

  /**
   * @see Oox_Lock_Interface::checkEnvironment()
   */
  public function checkEnvironment() {
    return TRUE;
  }
}
