<?php
// $Id: Interface.php,v 1.1.2.1 2011/01/04 18:23:27 pounard Exp $

/**
 * Simple semaphore based persistant locking framework. This interface defines
 * the lock backend API.
 * 
 * Locks are acquired on a name basis. Each acquired lock will raise exception
 * on new acquire attempts until it's being either pragmaticaly released, or
 * when its expiration time comes out of date.
 * 
 * For each lock, you can also give a token value. Any locking attempt of the
 * exact same lock with the exact same token will return sucessful.
 * 
 * If user is currently logged, each semaphore can store its uid. When done,
 * all current active locks will be released on user session destroy.
 * 
 * A default backend using a database table for locking is given, and used per
 * default.
 * 
 * @see Oox_Lock
 * @see oox_user()
 */
interface Oox_Lock_Interface
{
  const DEFAULT_GROUP = 'default';

  /**
   * Acquire a lock.
   * 
   * @param string $name
   *   Lock name.
   * @param int $lifetime = LIFETIME_DEFAULT
   *   (optional) Lock lifetime (in seconds).
   * @param string $group = Oox_Lock_Interface::DEFAULT_GROUP
   *   (optional) A group name for lock, this will allow to remove all locks for
   *   a certain group.
   * @param boolean $forUser = FALSE
   *   (optional) If set to TRUE, the current lock will be linked to the current
   *   logged in user (except for anonymous). This means that on session destroy
   *   if no other session is active, the lock will be removed.
   * @param string $token = NULL
   *   (optional) If set, set a token the lock. Tokens will allow more than one
   *   locking attempt on the same lock, using the same token. It can also allow
   *   lock sharing among users.
   * 
   * @throws LockAlreadyAcquiredException
   *   If lock already been acquired by another thread.
   */
  public function acquire($name, $lifetime = Oox_Lock_Interface::LIFETIME_DEFAULT, $group = Oox_Lock::DEFAULT_GROUP, $forUser = FALSE, $token = NULL);

  /**
   * Release a lock.
   * If lock is already free or does not exists, remains silent.
   * 
   * @param string $name
   *   Lock name.
   */
  public function release($name);

  /**
   * Get alive lock.
   * 
   * @param string $name
   *   Lock name.
   * @param string $group = NULL
   *   (optional) If set, will attempt to find by matching group.
   * @param int $uid = NULL
   *   (optional) If set, will attempt to find by matching uid.
   * @param string $token
   *   (optional) If set, will attempt to find by matching token.
   * 
   * @return object
   *   Lock database object if locked, FALSE else.
   */
  public function getLock($name, $group = NULL, $uid = NULL, $token = NULL);

  /**
   * Clear outdated locks.
   * 
   * Both of the following parameters can be used at once.
   * 
   * @param int $uid = 0
   *   (optional) If set, this will only wipe out all locks for the given
   *   user identifier.
   * @param string $group = NULL
   *   (optional) If set, this will wipe out all locks for the given group.
   * @param string $token = NULL
   *   (optional) If set, this will wipe out all locks for the given token.
   */
  public function wipeOutLocks($uid = 0, $group = NULL, $token = NULL);

  /**
   * Some cache backend can not run on some environment. This function will be
   * called at singleton instanciation time, if it returns FALSE, the singleton
   * instanciation will drop to the default backend as failsafe choice.
   * 
   * Remember that this check will done at each page request, it is your job to
   * ensure it won't be ressource intensive.
   * 
   * @return boolean
   */
  public function checkEnvironment();
}
