<?php

/**
 * Null object registry.
 * 
 * Whenever you need a dummy object because of a missing implementation, ask
 * it through this registry. If you do so, dummy objects will be shared among
 * all the code during execution time, which will save memory and CPU. These
 * objects will be used as flyweight objects.
 * 
 * <code>
 *   // You have this interface for your custom components.
 *   interface My_Module_Interface
 *   {
 *     public doSomething();
 *   }
 *   
 *   // You probably have some specific implementations such as this.
 *   class My_Module_Specific_Class implements My_Module_Interface
 *   {
 *     public doSomething() {
 *       something();
 *     }
 *   }
 * 
 *   // You can then define a dummy (neutral) implementation.
 *   class My_Module_Dummy 
 *     implements My_Module_Interface, Oox_NullObject_Interface
 *   {
 *     public doSomething() {}
 *   }
 *
 *   // Sometime, a caller might ask for a non existing implementation you
 *   // don't know nothing about, you could just let PHP crash:
 *   $class = 'Some_Unknown_Class';
 *   $myObject = new $class(); // Will fail with a WSOD
 *   
 *   // But, because the code could run in production mode, in front of real
 *   // end-users, you might want to give a dummy object instead. The execution
 *   // will remain transparent, instead of crashing it will just do nothing:
 *   $class = 'Some_Unknown_Class';
 *   if (class_exists($class) {
 *     $myObject = new $class();
 *   }
 *   else {
 *     $myObject = Oox_NullObject::getInstance('My_Module_Dummy');
 *   }
 * </code>
 * 
 * The NULL Object pattern has specific use case, one is, for example, NULL
 * date objects. There is no way letting some end-user data listing crash
 * because of wrong user input.
 * 
 * It could also be used by Views module, for unknown handlers. Instead of
 * using a NULL Object pattern, they just doesn't instanciate the unknown
 * class. Sometimes, you don't have this ease of control over your API where
 * the NULL Object pattern, tied to your object factories could generalize
 * this and avoid you having bad surprises at runtime.
 */
class Oox_NullObject
{
  /**
   * Already instanciated null objects.
   * 
   * @var array
   *   Array of Oox_NullObject_Interface instances.
   */
  private static $__instances = array();

  /**
   * Get null object singleton per class.
   * 
   * @param string $class
   *   Null object class.
   */
  public static function getInstance($class) {
    if (!isset(self::$__instances[$class])) {
      self::$__instances[$class] = new $class;
    }
    return self::$__instances[$class];
  }
}
