<?php
// $Id: Interface.php,v 1.1.2.1 2011/01/04 18:23:27 pounard Exp $

/**
 * Define an objet that can have a set of options.
 * 
 * The Oox_Registrable_Interface interface tells us the instances are objects
 * that can be loaded using a specific registry.
 * 
 * In all method signatures of this object, $name is a raw string that can
 * contain one or more Oox_Optionable_Interface::PATH_SEP characters. This gives
 * an options hierarchy. This option hierarchy will be implemented in default
 * implemention using a multidimensional array.
 */
interface Oox_Optionable_Interface extends Oox_Registrable_Interface
{
  /**
   * Separator for sub arrays in options. This tree handling will allow us
   * to deal only with numeric and string values inside the database value
   * field, and also keep track of parenting inside arrays.
   * 
   * @var char
   */
  const PATH_SEP = '\\';

  /**
   * Get options storage identifier.
   * 
   * @return int
   *   Can return NULL if object is not stored.
   */
  public function getOptionsIdentifier();

  /**
   * Set options storage identifier. This is an internal method for the
   * storage only, please do not use it.
   * 
   * @param int $oid
   *   Oox_Optionable storage identifier.
   * 
   * @return Oox_Optionable_Interface
   *   Self reference for chaining.
   */
  public function _setOptionsIdentifier($oid);

  /**
   * Get options count.
   * 
   * return int
   */
  public function getOptionsCount();

  /**
   * Set all instance specific options. This likely will be used only by
   * storages and registries, and such, and should never be called during
   * the objet normal life line.
   * 
   * @param array $options
   *   Oox_Optionable multidimensional array.
   * @param boolean $reset = FALSE
   *   (optional) If set to TRUE, new options will just erase the exisint one.
   *   Else it will merge nicely new options, erasing already set conflicting
   *   ones.
   * 
   * @return Oox_Optionable_Interface
   *   Self reference for chaining.
   */
  public function setOptions(array $options, $reset = FALSE);

  /**
   * Get specific options for an instance. 
   * 
   * @return array
   *   Options.
   */
  public function getOptions();

  /**
   * Shortcut that set a specific option.
   * 
   * @param string $name
   *   Option name.
   * @param mixed $value
   *   Option value.
   * 
   * @return Oox_Optionable_Interface
   *   Self reference for chaining.
   */
  public function setOption($name, $value);

  /**
   * Shortcut that get a specific option.
   * 
   * @param string $name
   *   Option name.
   * @param mixed $default = NULL
   *   (optional) Default value to return if option is not found.
   * 
   * @return mixed
   *   Option value, NULL if not found (or if the value is NULL).
   */
  public function getOption($name, $default = NULL);

  /**
   * Tells if option is set.
   * 
   * @return boolean
   *   TRUE if option is set (even if empty), FALSE else.
   */
  public function hasOption($name);

  /**
   * Unset the given option if exists.
   * 
   * @return Oox_Optionable_Interface
   *   Self reference for chaining.
   */
  public function removeOption($name);
}
