<?php

/**
 * Simple registry pattern implementation, using Drupal hook/cache systems
 * in order for module easily declare specific implementations.
 * 
 * In order to use it, the internal type must match the corresponding hook.
 * E.g. if you define the 'my_element' type, implementing modules must use
 * the hook_my_element() hook in order to expose their custom types.
 */
class Oox_Registry extends Oox_Registrable_Abstract
{
  /**
   * Per type singleton implementation.
   * 
   * @var array
   *   Full with Oox_Registry instances, or overriding classes.
   */
  private static $__instances = array();

  /**
   * Ugly, but working simple registry of registries implementation.
   * 
   * @var array.
   */
  private static $__registries;

  /**
   * Get instance of given type.
   * 
   * @param string $type
   *   Registry type.
   * @param boolean $reset = FALSE
   *   If set to TRUE, it will call the hook_oox_registry() again in
   *   order to rebuild the internal registry cache, if modules have been
   *   loaded after the first call.
   *
   * @return Oox_Registry
   *   Specialized implementation.
   * 
   * @throws Oox_Registry_Exception
   *   If registry type does not exists.
   */
  public static function getInstance($type, $reset = FALSE) {
    self::__loadCache($reset);
    // Lazzy load the registry when needed.
    if (!isset(self::$__instances[$type])) {
      $hook = NULL;
      $class = NULL;
      // Registries pseudo registry introspection.
      if (isset(self::$__registries[$type])) {
        // Find if there is a custom class given.
        if (isset(self::$__registries[$type]['class'])) {
          $class = self::$__registries[$type]['class'];
        }
        // Also check out for a given hook name.
        if (isset(self::$__registries[$type]['hook'])) {
          $hook = self::$__registries[$type]['hook'];
        }
        // Check for a name.
        if (isset(self::$__registries[$type]['name'])) {
          $name = self::$__registries[$type]['name'];
        }
        // Check for condition/properties.
        if (isset(self::$__registries[$type]['properties'])) {
          $properties = self::$__registries[$type]['properties'];
        }
      }
      else {
        throw new Oox_Registry_Exception("Registry " . $type . " does not exists");
      }
      // If not exists, then all we need is the classical object registry,
      // which does most of the job developers will need though.
      $registry = $class ? new $class($type) : new Oox_Registry($type);
      $registry->_setModule(self::$__registries[$type]['module']);
      if (isset($hook)) {
        $registry->_setHookName($hook);
      }
      if (isset($name)) {
        $registry->_setName($name);
      }
      if (isset($properties)) {
        $registry->_setProperties($properties);
      }
      self::$__instances[$type] = $registry;
    }
    return self::$__instances[$type];
  }

  /**
   * Load static registries cache.
   */
  private static function __loadCache($reset = FALSE) {
    if ($reset || !isset(self::$__registries)) {
      // FIXME: Handle at least some cache level here.
      $hook = 'oox_registry';
      self::$__registries = array();
      foreach (module_implements($hook) as $module) {
        foreach (module_invoke($module, $hook) as $type => $item) {
          $item['module'] = $module;
          self::$__registries[$type] = $item;
        }
      }
    }
  }

  /**
   * Get all registries descriptions.
   * 
   * @return array
   */
  public static function _getAllDescriptions() {
    self::__loadCache();
    // FIXME: This should have not to be done, because we don't do caching
    // we do this filtering on the less asked method to keep the global static
    // initialization performant in every day's life.
    $ret = array();
    foreach (self::$__registries as $type => &$item) {
      if (!empty($item)) {
        $ret[$type] = $item;
      }
    }
    return $ret;
  }

  /**
   * Main constructor. This is protected due to the singleton pattern.
   * 
   * @param string $type
   *   Registry type.
   */
  protected function __construct($type) {
    $this->_setType($type);
  }

  /**
   * Available properties for subset items selection.
   */
  protected $_properties = array();

  /**
   * Get available properties for conditional item instanciation.
   * 
   * @return array
   *   Key/value pairs, keys are properties names while values are properties
   *   human non-localized name, if set.
   *   If no human names are provided by the defining module, keys and values
   *   are equal.
   */
  public function getProperties() {
    return $this->_properties;
  }

  /**
   * Set properties.
   * 
   * @param array $properties
   *   Key/value pairs, keys are properties names while values are properties
   *   human non-localized name, if set. If no human names are being available,
   *   please provide a valued only array, where values are properties names.
   */
  protected function _setProperties(array $properties) {
    current($properties);
    // Determine which type of array is given.
    list($key, $value) = each($properties);
    if (is_numeric($key)) {
      $this->_properties = array();
      foreach ($properties as $value) {
        $this->_properties[$value] = $value;
      }
    }
    else {
      $this->_properties = $properties;
    }
  }

  /**
   * Internal types cache.
   * 
   * @var array
   */
  protected $_cache;

  /**
   * Internal properties cache.
   * 
   * @var array
   */
  protected $_propertyCache;

  /**
   * Set to private to force overriding classes to use the _getHookName()
   * method instead.
   * 
   * @var string
   */
  private $__hookName;

  /**
   * Pragmatically change hook name for this registry. This will be usefull
   * for registry instanciation time, if module owner did not want to define
   * a new class only to change that hook name.
   * 
   * @param string $hook
   *   Hook name.
   */
  protected function _setHookName($hook) {
    $this->__hookName = $hook;
  }

  /**
   * Get hook name and cache identifier, if you don't like it, you can
   * override it.
   * 
   * @return string
   */
  public function _getHookName() {
    if (!isset($this->__hookName)) {
      $this->__hookName = 'oox_registry_' . $this->_type;
    }
    return $this->__hookName;
  }

  /**
   * Registry name.
   * 
   * @var string
   */
  private $__name;

  /**
   * Set registry name.
   */
  protected final function _setName($name) {
    $this->__name = $name;
  }

  /**
   * Get registry name.
   */
  public final function getName() {
    if (!empty($this->__name)) {
      return $this->__name;
    }
    else {
      return $this->_type . " registry";
    }
  }

  /**
   * (Re)populate internal cache.
   * 
   * Set to public with the internal proctected prefix _ so other modules can
   * access it in order to force cache rebuild.
   * 
   * This will also produce a fast lookup properties array.
   */
  public function _rebuildItemCache() {
    $this->_cache = $this->_propertyCache = array();
    $hook = $this->_getHookName();

    // Handle custom item ordering.
    $weights = variable_get($hook . '_weight', array());

    // Call hook on each module that intends to provide items for this custom
    // registry.
    foreach (module_implements($hook) as $module) {
      foreach (module_invoke($module, $hook) as $type => $item) {
        // Allow overriding class to add its own checks.
        if ($this->_checkItem($type, $item, $module)) {

          // If all ok, then add our item to cache.
          $this->_cache[$type] = $item;
          // Handle custom item ordering.
          if (isset($weights[$type])) {
            $this->_cache[$type]['weight'] = $weights[$type]; 
          }
          else {
            $this->_cache[$type]['weight'] = 0;
          }

          // Object is valid, rebuild its property cache.
          if (isset($item['properties'])) {
            foreach ($item['properties'] as $key => $values) {
              // Ignore system unknown properties.
              if (!isset($this->_properties[$key])) {
                continue;
              }
              // Normalize values.
              if (!is_array($values)) {
                $values = array($values);
              }
              // Populate lookup cache.
              foreach ($values as $value) {
                $this->_propertyCache[$key][$value][] = $type;
              }
            }
          }
        }
      }
    }

    // Real sorting.
    uasort($this->_cache, '_oox_registry_compare');

    // Save the cached result.
    cache_set($hook, array($this->_cache, $this->_propertyCache), OOX_CACHE_TABLE);
  }

  /**
   * Check an item coming from a specific module hook implementation.
   * 
   * You are advised to override this method for specific implementations, don't
   * ever forget to call parent implementation.
   * 
   * @param string $type
   *   Element type name.
   * @param array &$item
   *   Raw item descriptive array from hook implementation.
   * @param string $module
   *   Original module name.
   * 
   * @return boolean
   *   TRUE if item is OK, false else.
   */
  protected function _checkItem($type, &$item, $module) {
    $item['module'] = $module;
    $item['type'] = $type;

    // Trigger some warnings.
    if (isset($item['file'])) {
      watchdog('oox', "Key 'file' in registry description is deprecated, use hook_oox_api() instead. Error triggered by module '" . $module . "' for item '" . $type . "'.", NULL, WATCHDOG_ALERT);
    }

    // If file exists, load it for further verifications.
    if (isset($path)) {
      $item['file'] = $path;
      require_once $item['file'];
    }

    // Check for defined class.
    if (!isset($item['class']) || !class_exists($item['class'])) {
      watchdog('oox', "Module " . $module . " define the class '" . $item['class'] . "' which does not exists, unregistering object type " . $type, NULL, WATCHDOG_ERROR);
      return FALSE;
    }
    return TRUE;
  }

  /**
   * Load cache from Drupal cache.
   */
  protected function _loadItemCache() {
    // Lazzy load the cache.
    if (!isset($this->_cache)) {
      $cid = $this->_getHookName();
      if ($cached = cache_get($cid, OOX_CACHE_TABLE)) {
        $this->_cache = $cached->data[0];
        $this->_propertyCache = $cached->data[1];
      }
      else {
        $this->_rebuildItemCache();
      }
    }
  }

  /**
   * Attempt to find item type depending on what is given.
   * 
   * @param mixed $item
   * 
   * @throws Oox_Exception
   *   If not found.
   */
  protected function _getItemType($item) {
    if (is_string($item)) {
      return $item;
    }
    else if ($item instanceof Oox_Registrable_Interface) {
      return $item->getType();
    }
    else if (is_object($item)) {
      $class = get_class($item);
      foreach ($this->getItemCache() as $type => $desc) {
        if (isset($desc['class']) && $class === $desc['class']) {
          return $type;
        }
      }
    }
    throw new Oox_Exception("Unknown or null object provided.");
  }

  /**
   * Get back the internal cache array.
   * 
   * @param $type = NULL
   *   (optional) If given, return data for given type only.
   * 
   * @return array
   *   Raw item descriptive arrays, given by modules with some modifications
   *   made by the _checkItem() method.
   * 
   * @throws Oox_Registry_Exception
   *   In case the provided type does not exists.
   */
  public function getItemCache($type = NULL) {
    $this->_loadItemCache();
    if ($type) {
      $type = $this->_getItemType($type);
      if (!isset($this->_cache[$type])) {
        throw new Oox_Registry_Exception("Element type " . $type . " is unknown for registry type " . $this->_type);
      }
      return $this->_cache[$type];
    }
    return $this->_cache;
  }

  /**
   * Get item type administrative name.
   * This is a helper for form developers.
   * 
   * @param string|Oox_Registrable_Interface $type
   *   Either a live instance of Oox_Registrable_Interface, either a string item type.
   * 
   * @throws Oox_Registry_Exception
   *   If instance or type is not handled by this registry.
   */
  public function getItemName($type) {
    $item = $this->getItemCache($this->_getItemType($type));
    return isset($item['name']) ? $item['name'] : $type . " item";
  }

  /**
   * Get item type administrative description.
   * This is a helper for form developers.
   * 
   * @param string|Oox_Registrable_Interface $type
   *   Either a live instance of Oox_Registrable_Interface, either a string item type.
   * 
   * @throws Oox_Registry_Exception
   *   If instance or type is not handled by this registry.
   */
  public function getItemDescription($type) {
    $item = $this->getItemCache($this->_getItemType($type));
    return isset($item['description']) ? $item['description'] : NULL;
  }

  /**
   * Returns the full range of supported item types.
   * 
   * FIXME: It might be valuable to do a static cache here?
   * 
   * @return array
   *   Keys are be types, values are the human readable names.
   */
  public function getItemTypes() {
    $ret = array();
    $this->_loadItemCache();
    foreach ($this->_cache as $type => $item) {
      $ret[$type] = isset($item['name']) ? $item['name'] : $type;
    }
    return $ret;
  }

  /**
   * Tell if given item type is supported. If you have only one check to do,
   * call this method, it will be faster than getting the full range then
   * check in it.
   * 
   * @return boolean
   */
  public final function isItemTypeSupported($type) {
    $this->_loadItemCache();
    return isset($this->_cache[$type]);
  }

  /**
   * Get one instance for each type registered.
   * 
   * @param Schema_Interface $options = NULL
   *   (optional) If given these options will be passed to each object while
   *   instanciating it
   * 
   * @return array
   *   Key/value pairs, keys are item type, and values are new object instances.
   */
  public function getAllItems(Schema_Interface $options = NULL) {
    $ret = array();
    $this->_loadItemCache();
    foreach ($this->_cache as $type => &$item) {
      $ret[$type] = $this->getItem($type, $options);
    }
    return $ret;
  }

  /**
   * Get one item for each type given in the array, in the same order.
   * 
   * @param array $types
   * @param Schema_Interface $options = NULL
   *   (optional) If given these options will be passed to each object while
   *   instanciating it.
   * 
   * @return array
   *   Array of new object instances.
   * 
   * @throws Oox_Registry_Exception
   */
  public function getItems(array $types, Schema_Interface $options = NULL) {
    $ret = array();
    foreach ($types as $type) {
      $ret[$type] = $this->getItem($type, $options);
    }
    return $ret;
  }

  /**
   * Get new element instance. You can override it, but be aware of cache load
   * if you do so.
   * 
   * @param string $type
   *   Element type.
   * @param Schema_Interface $options = NULL
   *   (optional) Object options.
   *
   * @return mixed
   *   Type instance.
   * 
   * @throws Oox_Registry_Exception
   */
  public function getItem($type, Schema_Interface $options = NULL) {
    $this->_loadItemCache();
  
    // Check type exists.
    if (!isset($this->_cache[$type])) {
      throw new Oox_Registry_Exception("Element type " . $type . " is unknown for registry type " . $this->_type);
    }

    // Create new instance and return element if all ok.
    if (isset($this->_cache[$type]['file'])) {
      require_once $this->_cache[$type]['file'];
    }

    $class = $this->_cache[$type]['class'];

    // This may happen, when installing or updating.
    if (!class_exists($class)) {
      throw new Oox_Registry_Exception("Early bootstrap, autoload function seems not be registered yet");
    }

    // Create new item.
    $item = new $class();

    // Populate information for the Oox_Registrable_Interface interface.
    // Some objects may not be registrables from here.
    if ($item instanceof Oox_Registrable_Interface) {
      $item
        ->_setModule($this->_cache[$type]['module'])
        ->_setType($type)
        ->_setRegistryType($this->_type);
    }
    if ($item instanceof Schema_Exportable_Interface && isset($options)) {
      $item
        ->restoreFromSchema($options);
    }

    return $item;
  }

  /**
   * Get all instances that matches the given conditions.
   * 
   * @param array $conditions
   *   Key/value pairs. Keys are condition name, fixed by specific registry
   *   definition while pairs are either scalar values, or arrays of scalar
   *   values for basic OR condition.
   * @param Schema_Interface $options = NULL
   *   (optional) If given these options will be passed to each object while
   *   instanciating it.
   * @param boolean $silentErrors = FALSE
   *   (optional) If set to TRUE, do not throw exceptions on errors.
   * 
   * @return array
   *   Array of instances.
   * 
   * @throws Oox_Registry_Exception
   *
   * @see hook_oox_registry()
   */
  public function getItemsByProperties(array $conditions = array(), Schema_Interface $options = NULL, $silentErrors = FALSE) {
    $types = NULL;

    foreach ($conditions as $key => $values) {
      $potential = array();

      // Check for existing property. Non existing properties should always
      // lead to a decent fail.
      if (!isset($this->_properties[$key])) {
        if (!$silentErrors) {
          throw new Oox_Registry_Exception("Property '" . $key . "' is unknown for object from registry '" . $this->_registryType . "'");
        }
        continue;
      }

      // Normalize values.
      if (!is_array($values)) {
        $values = array($values);
      }

      // Find out objects.
      foreach ($values as $value) {
        if (isset($this->_propertyCache[$key][$value])) {
          $potential += $this->_propertyCache[$key][$value];
        }
      }

      // Filter types with already known ones.
      // On the first iteration or if we have only one condition, the currently
      // found potential types are all the allowed ones. We won't filter. Once
      // we reached at least the second iteration, conditions are AND therefore
      // types that are not in both arrays should be rejected. 
      if (isset($types)) {
        $types = array_intersect($types, $potential);
      }
      else {
        $types = $potential;
      }

      // Whenever the types array is empty, it means that we did too many
      // conditions therefore we have no potential matches.
      if (empty($types)) {
        return array();
      }
    }

    // Instanciate items.
    return $this->getItems(array_unique($types, $options));
  }
}

/**
 * Small static helper for registry items sorting.
 */
function _oox_registry_compare($a, $b) {
  if ($a['weight'] == $b['weight']) {
    return strcmp($a['type'], $b['type']);
  }
  else {
    return $a['weight'] - $b['weight'];
  }
}
