<?php

/**
 * Uses a PHP array as stream source.
 * 
 * <code>
 *   // You have an array of stuff you fetched back from somewhere, and you
 *   // wish to use it as an object stream in order to integrate with other
 *   // with other pieces of API, then just:
 *   $array = something();
 *   
 *   // This will create the stream instance using the datatype node.
 *   // 'raw' datatype is the default.
 *   $stream = new ObjectStream_Browseable_Array($array, 'node');
 * 
 *   // Then you can use it whenever you want.
 * </code>
 * 
 * The setLimit() and setOffset() methods here maybe have no sense at all since
 * you can pass the array you like at __construct() time. While this doesn't
 * really make a lot of sense, you can still use it. It will provide you paging
 * for your own array.
 */
class ObjectStream_Browseable_Array
  extends Oox_Registrable_Abstract
  implements ObjectStream_Browseable_Interface
{
  /**
   * @var string
   */
  protected $_datatype = OBJECT_STREAM_DATATYPE_RAW;

  /**
   * Allow user to pragmatically change datatype.
   * 
   * @param string $datatype
   * 
   * @return ObjectStream_Browseable_Array
   */
  public function setDatatype($datatype) {
    $this->_datatype = $datatype;
    return $this;
  }

  /**
   * @see ObjectStream_Interface::getDatatype()
   */
  public function getDatatype() {
    return $this->_datatype;
  }

  /**
   * @var array
   */
  protected $_objects;

  /**
   * For Iterable implementation only.
   * 
   * @var int
   */
  protected $_pos = 0;

  /**
   * @var int
   */
  protected $_limit = ObjectStream_Browseable_Interface::NO_LIMIT;

  /**
   * @see ObjectStream_Browseable_Interface::setLimit()
   */
  public function setLimit($limit) {
    $this->_limit = $limit;
    return $this;
  }

  /**
   * @see ObjectStream_Browseable_Interface::getLimit()
   */
  public function getLimit() {
    return $this->_limit;
  }

  /**
   * @var int
   */
  protected $_offset = 0;

  /**
   * @see ObjectStream_Browseable_Interface::setOffset()
   */
  public function setOffset($offset) {
    $this->_offset = $offset;
    return $this;
  }

  /**
   * @see ObjectStream_Browseable_Interface::getOffset()
   */
  public function getOffset() {
    return $this->_offset;
  }

  /**
   * @see ObjectStream_Browseable_Interface::setRange()
   */
  public function setRange($offset = NULL, $limit = NULL) {
    if (isset($offset)) {
      $this->_offset = $offset;
    }
    if (isset($limit)) {
      $this->_limit = $limit;
    }
  }

  /**
   * @see Iterator::current()
   */
  public function current() {
    return current($this->_objects);
  }

  /**
   * @see Iterator::key()
   */
  public function key() {
    return key($this->_objects);
  }

  /**
   * @see Iterator::next()
   */
  public function next() {
    next($this->_objects);
    ++$this->_pos;
  }

  /**
   * @see Iterator::rewind()
   */
  public function rewind() {
    reset($this->_objects);
    $this->_pos = 0;
  }

  /**
   * @see Iterator::valid()
   */
  public function valid() {
    return $this->_pos < count($this->_objects);
  }

  /**
   * @see Countable::count()
   */
  public function count() {
    return count($this->_objects);
  }

  /**
   * Default constructor.
   */
  public function __construct($objects = array(), $datatype = OBJECT_STREAM_DATATYPE_RAW) {
    $this->_objects = $objects;
  }
}
