<?php

/**
 * This implementation will be able to fetch any kind of entity using the
 * EntityFieldQuery class as backend.
 * 
 * This implementation provides a static caching mecanism which allows you
 * to iterate over fetched objects more than once. Each time you modify any
 * of this object's properties, this cache is cleared properly and query is
 * run again.
 * 
 * Querying a non existing entity type will fail at query time, not before.
 * This is your responsability to ensure that you are actually querying an
 * existing entity type.
 * 
 * <code>
 *   // Basic node example.
 *   $stream = new ObjectStream_Entity_Query;
 *   $stream
 *     ->setEntityType('node')
 *     ->setRange(2, 10)
 *   foreach ($stream as $node) {
 *     // Do something.
 *   }
 *   
 *   We could even render those directly:
 *   return node_view_multiple(new ObjectStream_Entity_Query('node', 0, 10));
 * 
 *   // Basic user example.
 *   $stream = new ObjectStream_Entity_Query;
 *   $stream
 *     ->setEntityType('user')
 *     ->setRange(1, 16)
 *   foreach ($stream as $user) {
 *     // Do something.
 *   }
 * </code>
 */
class ObjectStream_Entity_Query
  extends Oox_Registrable_Abstract
  implements ObjectStream_Browseable_Interface, ObjectStream_Persistent_Interface
{
  /**
   * @var string
   */
  protected $_entityType;

  /**
   * Alias of the getEntityType() method.
   * 
   * @see ObjectStream_Interface::getDatatype()
   */
  public function getDatatype() {
    return $this->_entityType;
  }

  /**
   * Set entity type for querying.
   * 
   * This entity type will be exposed as stream's datatype.
   * 
   * @param string $entityType
   * 
   * @return ObjectStream_Entity
   */
  public function setEntityType($entityType) {
    $this->_entityType = $entityType;
    return $this;
  }

  /**
   * Get entity type.
   * 
   * @return string
   */
  public function getEntityType() {
    return $this->_entityType;
  }

  /**
   * @var int
   */
  protected $_limit = ObjectStream_Browseable_Interface::NO_LIMIT;

  /**
   * @see ObjectStream_Browseable_Interface::setLimit()
   */
  public function setLimit($limit) {
    $this->_limit = $limit;
    return $this;
  }

  /**
   * @see ObjectStream_Browseable_Interface::getLimit()
   */
  public function getLimit() {
    return $this->_limit;
  }

  /**
   * @var int
   */
  protected $_offset = 0;

  /**
   * @see ObjectStream_Browseable_Interface::setOffset()
   */
  public function setOffset($offset) {
    $this->_offset = $offset;
    return $this;
  }

  /**
   * @see ObjectStream_Browseable_Interface::getOffset()
   */
  public function getOffset() {
    return $this->_offset;
  }

  /**
   * @see ObjectStream_Browseable_Interface::setRange()
   */
  public function setRange($offset = NULL, $limit = NULL) {
    if (isset($offset)) {
      $this->_offset = $offset;
    }
    if (isset($limit)) {
      $this->_limit = $limit;
    }
  }

  /**
   * @var array
   */
  protected $_objects;

  /**
   * @var int
   */
  protected $_count;

  /**
   * @var int
   */
  protected $_pos = 0;

  /**
   * @var boolean
   */
  protected $_modified = FALSE;

  /**
   * @var ObjectStream_Entity_Query
   */
  protected $_query;

  /**
   * Prepare and the query.
   */
  protected function _initQuery() {
    if ($this->_modified || !isset($this->_query)) {
      // Clean up internals.
      $this->_modified = FALSE;

      // Prepare the field query.
      $this->_query = new ObjectStream_Entity_FieldQuery();
      $this->_query->entityCondition('entity_type', $this->getDatatype());

      /*
      foreach ($this->_options['filters'] as $filter) {
        $filter->alterQuery($this->_query);
      }
       */
    }
  }

  /**
   * @see Countable::count()
   */
  public function count() {
    if ($this->_modified || !isset($this->_count)) {
      $this->_initQuery();

      // Clone the actual query in order to get a count query.
      // Once again, this behavior differs from the actual SelectQuery from
      // DBTNG, which is quite a pain in the ass.
      $countQuery = clone $this->_query;
      $this->_count = $countQuery->count()->execute();
    }

    return $this->_count;
  }

  /**
   * Query the Drupal entity storage and fetch entities.
   * 
   * @return ObjectStream_Entity
   */
  public function executeQuery() {
    if ($this->_modified || !isset($this->_objects)) {
      $this->_initQuery();

      // Set limit and offset for paging requests.
      if ($this->_limit || $this->_offset) {
        $this->_query->range($this->_offset, $this->_limit);
      }

      // Use the entity system to query objects.
      $this->_objects = array();
      $entity = entity_get_controller($this->_entityType);
      $result = $this->_query->execute();

      // This one was tricky.
      $this->_objects = $entity->load(array_keys($result[$this->_entityType]));
    }
    return $this;
  }

  /**
   * @see Iterator::current()
   */
  public function current() {
    if (!isset($this->_objects)) {
      throw new Oox_Exception("Query did not run yet.");
    }
    return current($this->_objects);
  }

  /**
   * @see Iterator::key()
   */
  public function key() {
    if (!isset($this->_objects)) {
      throw new Oox_Exception("Query did not run yet.");
    }
    return key($this->_objects);
  }

  /**
   * @see Iterator::next()
   */
  public function next() {
    if (!isset($this->_objects)) {
      throw new Oox_Exception("Query did not run yet.");
    }
    next($this->_objects);
    return $this->_pos++;
  }

  /**
   * @see Iterator::rewind()
   */
  public function rewind() {
    if (!isset($this->_objects)) {
      $this->executeQuery();
    }
    reset($this->_objects);
    $this->_pos = 0;
  }

  /**
   * @see Iterator::valid()
   */
  public function valid() {
    if (!isset($this->_objects)) {
      throw new Oox_Exception("Query did not run yet.");
    }
    return $this->_pos < count($this->_objects);
  }

  /**
   * @see Vc_Backend_Interface::getFilterInterface()
   */
  /*
  public function getFilterInterface() {
    return 'VcEntity_Filter_Interface';
  }
   */

  /**
   * @see Schema_Exportable_Interface::getSchema()
   */
  public function getSchema() {
    return new Schema_Array(array(
      'entity_type' => $this->_entityType,
      'limit'       => $this->_limit,
      'offset'      => $this->_offset,
    ));
  }

  public function restoreFromSchema(Schema_Interface $config) {
    $this->_entityType = $config->get('entity_type');
    $this->_limit      = $config->get('limit');
    $this->_offset     = $config->get('offset');
  }

  /**
   * Default constructor.
   * 
   * @param string $entityType
   * @param int $offset = NULL
   * @param int $limit = NULL
   */
  public function __construct($entityType = NULL, $offset = NULL, $limit = NULL) {
    $this->_entityType = $entityType;
    $this->setRange($offset, $limit);
  }
}
