<?php

/**
 * Implementation of Schema_Array using a cache mecanism for store and retrieve
 * values.
 * 
 * This implementation will be super fast for reading operation, but will then
 * be quite heavy as soon as you do writes. This is due to the fact that
 * modificating the current object will need the full value array to be loaded
 * in memory to ensure all checks are done all right.
 */
abstract class Schema_Cached_Array extends Schema_Array implements Schema_Cached_Interface
{
  /**
   * @var Schema_Reader_Interface
   */
  protected $_reader = FALSE;

  /**
   * @var bool
   */
  protected $_updated = FALSE;

  /**
   * Current keys cache.
   * 
   * @var array
   */
  protected $_cache;

  /**
   * @see Schema_Array::get()
   */
  public function get($name) {
    if (!array_key_exists($name, $this->_cache)) {
      $this->_populateFromReader();
      $this->_cache[$name] = parent::get($name);
    }
    return $this->_cache[$name];
  }

  /**
   * @see Schema_Array::set()
   */
  public function set($name, $value) {
    $this->_populateFromReader();
    parent::set($name, $value);
    $this->_updated = TRUE;
    $this->_cache[$name] = $value;
  }

  /**
   * @see Schema_Array::getSection()
   */
  public function getSection($name) {
    $this->_populateFromReader();
    // This wont set the cache, because iterating over the section would
    // require us to have the right interface for it.
    return parent::getSection($name);
  }

  /**
   * @see Schema_Array::setSection()
   */
  public function setSection($name, Schema_Readonly_Interface $section) {
    $this->_populateFromReader();
    parent::setSection($name, $section);
    // Clear all cache, we don't have any choice here. The section path could
    // not exist therefore its children could.
    $this->cacheClear();
  }

  /**
   * @see Schema_Array::merge()
   */
  public function merge(Schema_Readonly_Interface $schema, $override = FALSE) {
    $this->_populateFromReader();
    parent::merge($schema, $override);
  }

  /**
   * @see Schema_Array::toArray()
   */
  public function toArray() {
    $this->_populateFromReader();
    return parent::toArray();
  }

  /**
   * Count won't give a right value until we really count root level data.
   * 
   * @see Countable::count()
   */
  public function count() {
    $this->_populateFromReader();
    return parent::count();
  }

  /**
   * Doing a foreach over this object needs him to populate all data.
   * 
   * @see Schema_Array::rewind()
   */
  public function rewind() {
    $this->_populateFromReader();
    parent::rewind();
  }

  /**
   * @var string
   */
  protected $_cid;

  /**
   * @var string
   */
  protected $_bin;

  /**
   * Populate data from reader, if set.
   */
  protected function _populateFromReader() {
    if (!isset($this->_data)) {
      $this->_data = array();
    }
    if (isset($this->_reader) && !isset($this->_data)) {
      // Override set to FALSE to ensure we won't erase local modifications
      // with potentially outdated ones from the reader.
      parent::merge($this->_reader->read(), FALSE);
    }
  }

  /**
   * @see Schema_Cached_Interface::cacheLoad()
   */
  public function cacheLoad() {
    if ($cached = cache_get($this->_cid, $this->_bin)) {
      $this->_cache = $cached->data;
    }
    else {
      $this->_populateFromReader();
      $this->_cache = array();
    }
    $this->_updated = FALSE;
  }

  /**
   * @see Schema_Cached_Interface::cacheSave()
   */
  public function cacheSave($force = TRUE) {
    if ($this->_updated || $force) {
      cache_set($this->_cid, $this->_cache, $this->_bin);
    }
    $this->_updated = FALSE;
  }

  /**
   * @see Schema_Cached_Interface::cacheClear()
   */
  public function cacheClear() {
    $this->_cache = array();
    cache_clear_all($this->_cid, $this->_bin);
    $this->_updated = FALSE;
    $this->cacheLoad();
  }

  /**
   * Base constructor.
   * 
   * @param string $cid
   *   Cache identifier for cache storage.
   * @param string $bin
   *   Cache bin name for cache storage.
   * @param Schema_Reader_Interface $reader = NULL
   *   (optional) If set, and if current array is not set, use this reader to
   *   get values from. This allows lazzy loading of values.
   *   It will be used mainly during merge(), getSection(), setSection(), and
   *   clone() method calls, or while initing the object without any cache set.
   *   When using this object only for reading, and if it already has been read
   *   once, nothing will be done with the reader at all and no I/O will be made
   *   except the one cache load.
   * @param bool $strict = TRUE
   *   If set to TRUE, this object will raise exception when reaching non
   *   existing pathes.
   */
  public function __construct($cid, $bin, Schema_Reader_Interface $reader = NULL, $strict = TRUE) {
    $this->_readonly = FALSE;
    $this->_strict = $strict;
    $this->_cid = $cid;
    $this->_bin = $bin;
    // Set NULL to data, this will help us controlling weither or not we
    // actually loaded the full configuration.
    $this->_data = NULL;
    $this->cacheLoad();
  }
}
