<?php

/**
 * Variables like implementation.
 * 
 * This particular object implements the Active Record pattern, each data
 * modification will be succeeded by a database update.
 * 
 * This implementation is über fast for read and override operations, but will
 * be slower when used for writing.
 * 
 * Whenever you merge or export this object, it will be incredibly slow and
 * will trigger off any cache update, these operations needs to load the full
 * dataset, so it would be an error to store it completely it would then be
 * a huge and inefficient cache entry.
 */
class Schema_System implements Schema_Cached_Interface, Schema_Active_Interface
{
  /**
   * Table name where to fetch and save raw and uncached values.
   */
  const DB_TABLE = 'oox_schema_system';

  /**
   * Cache identifier for saving this data.
   */
  const CACHE_CID = 'oox_schema_system';

  /**
   * Cache bin where where to save this data.
   */
  const CACHE_BIN = 'cache_bootstrap';

  /**
   * Set when all data has been forced to be loaded. While this data has been
   * loaded, cache won't reflect the normal usage anymore, so we won't save
   * cache from there.
   * 
   * @var bool
   */
  protected $_forced = FALSE;

  /**
   * @var bool
   */
  protected $_updated = FALSE;

  /**
   * @var bool
   */
  protected $_loaded = FALSE;

  /**
   * Current object cache state.
   * 
   * @var array
   */
  protected $_cache;

  /**
   * @see Schema_Readonly_Interface::get()
   */
  public function get($name) {
    if (!array_key_exists($name, $this->_cache)) {
      $this->_cache[$name] = unserialize(db_select(self::DB_TABLE, 't')
        ->fields('t', array('value'))
        ->condition('name', $name)
        ->execute()
        ->fetchField());
      $this->_updated = TRUE;
    }
    // Cache should mostly be updated only at get() time, this because only
    // really being used variable are really useful in there.
    return $this->_cache[$name];
  }

  /**
   * We won't update cache on set() operation. If the variable is never being
   * used by any module at any time, no need for us to store it into cache.
   * 
   * @see Schema_Interface::set()
   */
  public function set($name, $value) {
    db_insert(self::DB_TABLE)
      ->fields(array('name', 'value'), array($name, serialize($value)))
      ->execute();
    // Un-comment this for more aggressive cache update, it may cache unused
    // variable though.
    // $this->_cache[$name] = $value;
    // $this->_updated = TRUE;
  }

  /**
   * @see Schema_Readonly_Interface::getSection()
   */
  public function getSection($name) {
    if (!Schema_Helper::isPathValid($name, FALSE)) {
      throw new Schema_Exception("'" . $name . "' is an invalid path.");
    }
    return new Schema_Array($this->_sectionToArray($name));
  }

  /**
   * This should be implemented like merge().
   * 
   * @see Schema_Interface::setSection()
   */
  public function setSection($name, Schema_Readonly_Interface $section) {
    if (!Schema_Helper::isPathValid($name, FALSE)) {
      throw new Schema_Exception("'" . $name . "' is an invalid path.");
    }
    // Re-use existing code.
    $this->merge($section, TRUE, TRUE, $name . '.');
  }

  /**
   * @see Schema_Active_Interface::merge()
   */
  public function merge(Schema_Readonly_Interface $schema, $override = FALSE, $prefix = '') {
    // Load data only if necessary.
    if (!$override) {
      $this->_populateAll();
    }
    foreach ($schema as $key => $value) {
      $path = $prefix . $key;

      if ($value instanceof Schema_Readonly_Interface) {
        $this->merge($value, TRUE, $path . '.');
      }
      else if ($override || !isset($this->_cache[$path])) {
        $this->set($path, $value);
      }
    }
  }

  /**
   * @see Schema_Active_Interface::remove()
   */
  public function remove($name) {
    db_delete(self::DB_TABLE)->condition('name', $name)->execute();
    // Get rid of the cache, it's now outdated.
    $this->cacheClear();
  }

  /**
   * @see Schema_Active_Interface::removeAll()
   */
  public function removeAll(Schema_Readonly_Interface $schema, $prefix = '') {
    foreach ($schema as $key => $value) {
      $path = $prefix . $key;

      if ($value instanceof Schema_Readonly_Interface) {
        $this->removeAll($value, $path . '.');
      }
      else {
        $this->remove($path);
      }
    }
    // Do not update cache with temporary runtime overrides.
    $this->_forced = TRUE;
  }

  /**
   * @see Schema_Active_Interface::override()
   */
  public function override(Schema_Readonly_Interface $schema, $prefix = '') {
    foreach ($schema as $key => $value) {
      $path = $prefix . $key;

      if ($value instanceof Schema_Readonly_Interface) {
        $this->override($value, $path . '.');
      }
      else {
        $this->_cache[$path] = $value;
      }
    }
    // Do not update cache with temporary runtime overrides.
    $this->_forced = TRUE;
  }

  /**
   * In some rare case we might want to get all existing data.
   * 
   * This seems a bit overloaded but it might help. Once this object has loaded
   * all, it might be good to avoid setting its cache, so we are going to tell
   * it not to.
   */
  protected function _populateAll() {
    // Avoid multiple load once we got here.
    if (!$this->_loaded) {
      $values = db_select(self::DB_TABLE, 't')
        ->fields('t', array('name', 'value'))
        ->execute()
        ->fetchAll();
      foreach ($values as $value) {
        $this->_cache[$value->name] = unserialize($value->value); 
      }
      // See this property documentation for details.
      $this->_forced = $this->_loaded = TRUE;
    }
  }

  /**
   * @see Schema_Cached_Interface::cacheLoad()
   */
  public function cacheLoad() {
    if ($cached = cache_get(self::CACHE_CID, self::CACHE_BIN)) {
      $this->_cache = $cached->data;
    }
    else {
      $this->_cache = array();
    }
    $this->_updated = FALSE;
  }

  /**
   * @see Schema_Cached_Interface::cacheSave()
   */
  public function cacheSave($force = TRUE) {
    // See the property PHP documentation for details.
    if ($this->_forced) {
      return;
    }
    if ($this->_updated || $force) {
      cache_set(self::CACHE_CID, $this->_cache, self::CACHE_BIN);
    }
    $this->_updated = FALSE;
  }

  /**
   * @see Schema_Cached_Interface::cacheClear()
   */
  public function cacheClear() {
    $this->_cache = array();
    cache_clear_all(self::CACHE_CID, self::CACHE_BIN);
    $this->_updated = FALSE;
  }

  /**
   * Same as toArray() except it will return the given section. The algorithm
   * is almost the same, the one exception is the name substring test.
   * 
   * @return array
   */
  protected function _sectionToArray($sectionName) {
    $array = array();
    $this->_populateAll();
    $len = strlen($sectionName);
    foreach ($this->_cache as $name => $value) {
      if (substr($name, 0, $len) != $sectionName) {
        continue;
      }
      $path = explode('.', $name);
      $current =& $array;
      while ($key = array_shift($path)) {
        if (!isset($current[$key])) {
          $current[$key] = array();
        }
        $current =& $current[$key];
      }
      $current = $value;
    }
    return $array;
  }

  /**
   * This is probably a faster implementation than the array one, because it's
   * de-recursified. The only wrong thing here is we maybe manipulate too many
   * strings.
   * 
   * @see Schema_Readonly_Interface::toArray()
   */
  public function toArray() {
    $array = array();
    $this->_populateAll();
    foreach ($this->_cache as $name => $value) {
      $path = explode('.', $name);
      $current =& $array;
      while ($key = array_shift($path)) {
        if (!isset($current[$key])) {
          $current[$key] = array();
        }
        $current =& $current[$key];
      }
      $current = $value;
    }
    return $array;
  }

  /**
   * Default constructor, loads the initial cache if exists.
   */
  public function __construct() {
    $this->cacheLoad();
  }
}
