<?php

/**
 * @file
 * Oox module hooks definition and documentation.
 */

/**
 * Define a custom registry. Whatever happens here, if you only need the default
 * Oox_Registry implementation, you are allowed not to define this hook.
 * 
 * Be gentle with storages, their definition are not cached (yet), use the
 * default implementation when you can.
 * 
 * In the registry definition, all keys are optional. You can even not define
 * your registry, in which case it will be spawned using the default
 * implementation when asked.
 * 
 * Notice that if you don't define it in your module, you won't be able to
 * access its options in the "registries" administration screen. All usage will
 * be dynamic, at run time.
 * 
 * Options are:
 *   - 'class' : Class name that represent the registry. For some specific
 *     purpose, you can override the Oox_Registry class. Not that if you
 *     specify a class here, the Oox_Registry override will be checked.
 *     If the base implementation suits you, do not specify the class name.
 *   - 'hook' : If defined, the hook used in order to define specific objet
 *     classes will be the given value. If you don't specify it, it will be the
 *     default hook_oox_registry_TYPE(), where TYPE is the items array
 *     key. Wheither or not you override the hook name, the hook signature will
 *     remains the same.
 *   - properties: array
 *     (optional) An array of key/value pairs. Values are properties internal
 *     names where values are their human name. If you don't want to name your
 *     properties you can pass a single value'd array instead where keys are
 *     numeric values.
 *     Properties are meant to allow basic filtering when fetching more than
 *     once instance from a registry.
 *     Propertie values for later usage will always be scalar values. Non scalar
 *     values will crash the system or raise exceptions.
 *     This property here only declares potential filters, but won't affect
 *     runtime at all as longer as modules doesn't use it.
 *     Each specific component must implement a properties array if it want
 *     to benefit from this feature, see hook_oox_registry_TYPE() definition.
 * 
 * Example of code using a registry:
 * 
 * <code>
 *   // Fetch the registry.
 *   $registry = oox_registry_get('mymodule_registry');
 *   
 *   // Fetch a single instance.
 *   $object = $registry->getItem('mymodule_TYPE_item');
 *   // Fetch all known objects instances:
 *   $objects = $registry->getAllItems();
 * 
 *   // The same as the two above, but using configuration object for
 *   // specific implementations initialization:
 *   // Create the configuration object.
 *   $config = new Schema_Array(array(
 *     'option_a' => 'value a',
 *     'option_b' => 'value b',
 *   ));
 *   $object = $registry->getItem('mymodule_TYPE_item', $config);
 *   // Note that all objects will be inited with the same options here.
 *   $objects = $registry->getAllItems($config);
 * 
 *   // Fetch all instances that declare the 'nice' property to 'yes':
 *   $objects = $registry->getItemsByProperties(array(
 *     'nice' => 'yes',
 *   ));
 *
 *   // Fetch all instances that declare the 'nice' property to 'yes'
 *   // or to 'maybe':
 *   $objects = $registry->getItemsByProperties(array(
 *     'nice' => array('yes', 'maybe'),
 *   ));
 *
 *   // Fetch all instances that declare the 'nice' property to 'yes'
 *   // or 'maybe' but with the property 'beautiful' to 'no', using the
 *   // upper configuration object for new instances:
 *   $properties = array(
 *     'nice' => array('yes', 'maybe'),
 *     'beautiful' => 'no',
 *   );
 *   $objects = $registry->getItemsByProperties($properties, $config);
 * </code>
 * 
 * @see hook_oox_registry_TYPE()
 */
function hook_oox_registry() {
  $items[] = array();
  // Key is the registry type.
  $items['mymodule_registry'] = array(
    'class' => 'MyModuleRegistry',
    'file' => drupal_get_path('module', 'mymodule') . '/lib/MyModuleRegistry.php',
    'hook' => 'mymodule_items',
  );
  return $items;
}

/**
 * Define available items for the given registry.
 * 
 * TYPE refers to the registry type. If a registry named 'foo_bar' is defined
 * then this hook name will be: hook_oox_registry_foo_bar().
 * 
 * Sometimes, the registry can define its own hook (see the hook_oox_registry()
 * definition), then this hook will be same except it will then be named:
 * hook_REGISTRY_HOOK_ENTRY().
 */
function hook_oox_registry_TYPE() {
  $items[] = array();
  // There is no naming convention at all, just be aware of potential conflicts
  // between modules.
  $items['mymodule_TYPE_item'] = array(
    // Same note here, no forced naming convention.
    'class' => 'MyModuleTYPEItem',
    // Name is optional, but is highly recommended for later UX usage. It must
    // be written in good english, not localized from here since this result is
    // highly likely to be cached somewhere. Specific UX have the responsability
    // of calling themselves the t() function.
    'name' => "Some name",
  );
  return $items;
}
